#!/bin/bash

VERSION="2010-11-11"
# xarf-ssh-reporter.sh $VERSION Sven Uebelacker <software@dfn-cert.de>, DFN-CERT Services GmbH
# License: see attached LICENSE file

# loading config file, has to be in script directory or in /etc
CONFIG_FILE="xarf-ssh-reporter.conf"
SCRIPT_EXEC_PATH=${0%/*}
if [ -r "${SCRIPT_EXEC_PATH}/${CONFIG_FILE}" ]
then	. "${SCRIPT_EXEC_PATH}/${CONFIG_FILE}"
elif	[ -r "/etc/${CONFIG_FILE}" ]
then	. "/etc/${CONFIG_FILE}"
else	echo "error: could not load config file \"${CONFIG_FILE}\" neither in /etc nor in ${SCRIPT_EXEC_PATH}"
	exit 1
fi

########################
### begin self-check ###
########################
# check if bash version >= 4
if [ "${BASH_VERSINFO}" -lt 4 ]
then	echo "error: Your bash version is too old. You need at least bash version 4.0, you are using version ${BASH_VERSION}."
	exit 1
fi
# check if all logfiles are readable (-r also checks setfacl rights!)
notreadablelogfile=""
for logf in ${logfiles}
do	if [ ! -r "${logf}" ]
	then	notreadablelogfile+="${logf} "
	fi
done
if [ -n "${notreadablelogfile}" ]
then	echo "error: Can't get read access for following logfiles: ${notreadablelogfile}"
	exit 1
fi
# check if working_directory exists
if [ ! -d "${working_directory}" ]
then	echo "error: working directory \"${working_directory}\" does not exist."
	exit 1
fi
######################
### end self-check ###
######################

######################
### begin binaries ###
######################
BIN_SEARCH_LIST="GREP_BIN=grep EGREP_BIN=egrep WC_BIN=wc HEAD_BIN=head TAIL_BIN=tail CUT_BIN=cut SENDMAIL_BIN=sendmail TR_BIN=tr SORT_BIN=sort MD5SUM_BIN=md5sum"
missing_prog=""
for prog in ${BIN_SEARCH_LIST}
do      varname=${prog//=*}
        progname=${prog//*=}
        w=$( which ${progname} 2>/dev/null )
        if [ -z "${w}" ]
        then    missing_prog+="${progname} "
        else    eval ${varname}="${w}"
        fi
done
if [ -n "${missing_prog}" ]
then    echo "error: Can't find the following programs: ${missing_prog}"
        echo "error: \$PATH was \"${PATH}\""
        exit 1
fi
####################
### end binaries ###
####################

##################
### begin init ###
##################
SCRIPT_NAME=${0##*/}
declare -A iphash
declare -A lastparsedlogline
declare -A loglinecount
declare -A ipreportid
IP_MATCH_REGEXP='([0-9]{1,3}\.[0-9]{1,3}\.[0-9]{1,3}\.[0-9]{1,3})'
MIME_BOUNDARY="Abuse-"$(date|${MD5SUM_BIN}|${CUT_BIN} -d" " -f1)
################
### end init ###
################

#######################
### begin functions ###
#######################
ipv4_cidr_match() {
	# does ip address lies in the given CIDR defined subnet?
	# first arg: ipv4 address
	# second arg: CIDR
	# return 0=True, 1=False
	if [ -n "$1" -a -n "$2" ]
	then	ip="$1"
		cidr="$2"
		if [[ "${ip}" =~ ^([0-9]{1,3})\.([0-9]{1,3})\.([0-9]{1,3})\.([0-9]{1,3})$ ]]
		then	ipint=$(( ${BASH_REMATCH[1]}*256*256*256 + ${BASH_REMATCH[2]}*256*256 + ${BASH_REMATCH[3]}*256 + ${BASH_REMATCH[4]} ))
			if [[ "${cidr}" =~ ^([0-9]{1,3})\.([0-9]{1,3})\.([0-9]{1,3})\.([0-9]{1,3})/([0-9]{1,2})$ ]]
			then	cidr1=${BASH_REMATCH[1]}; cidr2=${BASH_REMATCH[2]}
				cidr3=${BASH_REMATCH[3]}; cidr4=${BASH_REMATCH[4]}
				mask=${BASH_REMATCH[5]}
				cidrlowipint=$(( ${cidr1}*256*256*256 + ${cidr2}*256*256 + ${cidr3}*256 + ${cidr4} ))
				iprange=$(( 2**(32-${mask})-1 ))
				cidrhighipint=$(( $iprange + $cidrlowipint ))
				if [ ${cidrlowipint} -le ${ipint} -a ${ipint} -le ${cidrhighipint} ]
				then return 0
				else return 1
				fi
			fi
		fi
	fi
	echo "error! usage: ipv4_cidr_match <ipv4> <cidr>"
}

ipv4_cidr_list_match() {
	# does ip address lies in the given CIDR defined subnet list?
	# first arg: ipv4 address
	# second arg: list of CIDRs
	# return 0=True, 1=False
	if [ -n "$1" -a -n "$2" ]
	then	ip="$1"
		# $2 - $N
		cidrlist=$(echo "$@" | ${CUT_BIN} -d" " -f2-)
		ret=1
		for cidr in ${cidrlist}
		do	if ipv4_cidr_match $ip $cidr
			then	ret=0
			fi
		done
		return ${ret}
	fi
	echo "error! usage: ipv4_cidr_list_match <ipv4> <cidr1> [<cidr2> ..]"
}

apply_cidr_blacklist() {
	# delete ip addresses which are in cidr blacklist
	# no args, uses global vars
	# no return value
	if [ -n "${cidr_blacklist}" ]
	then	for ip in ${!iphash[@]}
		do	if ipv4_cidr_list_match "${ip}" ${cidr_blacklist}
			then	unset iphash[${ip}]
			fi
		done
	fi
}

read_lastloglines() {
	# read lastloglines file to get informations about last processed loglines
	# no args, uses global vars
	# no return value
	[ ! -r "${lastloglinefile}" ] && : > "${lastloglinefile}"
	while read filename
	do	read logline
		# "[] escapen in logline for a save grep later
		temp=${logline//"/\\"}
		logline=${temp//[/\\[}
		temp=${logline//]/\\]}
		logline="${temp}"
		if [ -r "${filename}" ]
		then	lastparsedlogline[${filename}]="${logline}"
		fi
	done < "${lastloglinefile}"
}

write_lastloglines() {
	# write back the last parsed logline in the lastloglines file
	# no args, uses global vars
	# no return value
	for logfile in ${!lastparsedlogline[@]}
	do	echo "${logfile}"
		echo "${lastparsedlogline[${logfile}]}"
	done > "${lastloglinefile}"
}

read_and_parse_logfiles() {
	# open logfiles one by one and filter interesting logline in one big
	# templogfile; if there's an existing matching lastlogline start there
	# no args, uses global vars
	# no return value
	# create temporal file for interesting loglines:
	TEMPLOGFILE=$( mktemp --tmpdir="${TEMPDIR}" "${TEMPLOGFILETEMPLATE}" )

	# extract interesting loglines into tempfile
	for logfile in ${logfiles}
	do	if [[ "${remember_last_loglines}"x == "yes"x ]]
		then	# to do: does logfilename exist in hash lastparsedlogline?
			if ${GREP_BIN} "${lastparsedlogline[${logfile}]}" "${logfile}" &>/dev/null
			then	# logfile consists of a lastparsedlogline
				# => start parsing after lastparsedlogline
				# to be sure we get all lines after grep -A we use wc -l as argument
				wcl=$(${WC_BIN} -l "${logfile}" | ${CUT_BIN} -d" " -f1)
				${GREP_BIN} -A ${wcl} "${lastparsedlogline[${logfile}]}" "${logfile}" \
					| ${TAIL_BIN} -n+2 \
					| ${EGREP_BIN} "${egrep_regexp}" >> "${TEMPLOGFILE}"
			else	# logfile has no matching lastparsedlogline => parse all
				${EGREP_BIN} "${egrep_regexp}" "${logfile}" >> "${TEMPLOGFILE}"
			fi
			# remember new lastparsedlogline
			# to do: maybe there are new loglines since "${GREP_BIN} -A ${wcl}...", is there a fix?
			lastparsedlogline[${logfile}]=$(${TAIL_BIN} -n 1 "${logfile}")
		else	# we don't want to remember the last loglines
			${EGREP_BIN} "${egrep_regexp}" "${logfile}" >> "${TEMPLOGFILE}"
		fi
	done
}

beautify_logline_date() {
	# make the loglines more readable and better sortable by modifying the datetime format
	# one arg: one logline
	# returned value on stdout
	# output: beautified logline if first three columns can specify datetime
	# unless print original logline
	# i.e.: "Aug 14 16:05:06" -> "2010-08-14 16:05:06 +0200"
	# result is again a date time format of three columns
	if [ -n "$1" ]
	then	# substitute double spaces with one space
		line="${1//  / }"
		logdate=$(echo -e "${line}" | ${CUT_BIN} -d" " -f1-3)
		logrest=$(echo -e "${line}" | ${CUT_BIN} -d" " -f4-)
		if date +%F" "%H:%M:%S" "%z --date="${logdate}" &> /dev/null
		then	monthdaytimenow=$(date +%m%d%H%M%S)
			monthdaytimelog=$(date +%m%d%H%M%S --date="${logdate}")
			# standard syslog datetime lacks year, therefore if day/month/time is
			# after actual month/day/time => do a year-1
			if [ ${monthdaytimenow} -lt ${monthdaytimelog} ]
			then	year=$(date +%Y)
				year=$(( ${year} - 1 ))
				date +%F" "%H:%M:%S" "%z --date="${logdate} ${year}" | ${TR_BIN} -d "\n"
			else	date +%F" "%H:%M:%S" "%z --date="${logdate}" | ${TR_BIN} -d "\n"
			fi
			echo " ${logrest}"
			# to do: remember if conversion worked for later sorting/stripping
		else	echo "${line}"
		fi
	else	echo "ERROR: you must specify an logline as argument for beautify_logline_date."
		exit 1
	fi
}

build_iphash() {
	# build up associative array with ip addresses as keys and loglines as values
	# for sorting and threshold purpose
	# no args, uses global vars
	# no return value
	while read logline
	do	# do we have an ip address here in this logline?
		if [[ "${logline}" =~ ${IP_MATCH_REGEXP} ]]
		then	ip="${BASH_REMATCH[1]}"
			if [[ -z "${iphash[${ip}]}" ]]
			then	iphash[${ip}]=$(beautify_logline_date "${logline}")
			else	iphash[${ip}]+="\n"
				iphash[${ip}]+=$(beautify_logline_date "${logline}")
			fi
		fi
	done < "${TEMPLOGFILE}"

	# delete temporal logfile
	[ -f "$TEMPLOGFILE" ] && rm "$TEMPLOGFILE"
}

apply_threshold() {
	# use a given threshold to filter reportable ip addresses
	# delete ip addresses which have too few loglines
	# no args, uses global iphash
	# no return value
	for ip in ${!iphash[@]}
	do	linecount=$(echo -e "${iphash[$ip]}" | ${WC_BIN} -l)
		if [[ ${threshold} -gt ${linecount} ]]
		then	unset iphash[${ip}]
		else	loglinecount[${ip}]=${linecount}
		fi
	done
}

sort_and_limit_reportable_loglines() {
	# it will extract reportable_logline_limit/2 beginning and tailing loglines if over limit
	# if amount of reported loglines is less than reportable_logline_limit, report every line
	# loglines are used to be proof and not parsed later automatically
	# no args, uses global iphash
	# no return value
	if [[ -n "${reportable_logline_limit}" && ${reportable_logline_limit} -gt 0 ]]
	then	headtailcount=$(( ${reportable_logline_limit} / 2 ))
		for ip in ${!iphash[@]}
		do	if [[ ${loglinecount[${ip}]} -gt ${reportable_logline_limit} ]]
			then	newloglines=$( echo -e "${iphash[${ip}]}" | ${SORT_BIN} | ${HEAD_BIN} -n ${headtailcount} )
				newloglines+="\n-- MARK --\n"
				newloglines+=$( echo -e "${iphash[${ip}]}" | ${SORT_BIN} | ${TAIL_BIN} -n ${headtailcount} )
				iphash[${ip}]="${newloglines}"
			else	newloglines=$( echo -e "${iphash[${ip}]}" | ${SORT_BIN} )
				iphash[${ip}]="${newloglines}"
			fi
		done
	fi
}

anonymize_log_content() {
	# searches all loglines of iphash for the string "$anonymize_log_string"
	# and replaces it with 6 $anonymize_log_with_character
	# no args, uses global vars
	# no return value
	if [ -n "${anonymize_log_string}" ]
	then	anonstring="${anonymize_log_with_character}"
		for (( c=2; c<=6; c++ ))
		do	anonstring+="${anonymize_log_with_character}"
		done
		# because bash4 variable substitution can't handle underscore
		anonymizelogstring="${anonymize_log_string}"
		for ip in ${!iphash[@]}
		do	tempvalue=${iphash[${ip}]//${anonymizelogstring}/${anonstring}}
			iphash[${ip}]="${tempvalue}"
		done
	fi
}

build_report_mail_header_xarf_format() {
	# generate the mail header for a x-arf report
	# one optional arg with ip address
	# returned value on stdout
	# X-ARF is a mime type report format, one email per ip
	# http://www.x-arf.org/specification.html
	echo "From: ${from_email_address}"
	echo "To: ${report_email_address}"
	echo "Reply-To: ${replyto_email_address}"
	echo "X-Data-Format: ${export_format}"
	echo "Organisation: ${data_supplier_organisation}"
	echo "X-System-Id: ${system_id}"
	echo "X-Script-Version: ${VERSION}"
	echo "X-Script-Name: ${SCRIPT_NAME}"
	echo "X-Mailer: ${SCRIPT_NAME} ${VERSION}"
	echo "X-ARF: yes"
	echo "Auto-Submitted: auto-generated"
	datefield=$(date +%F)
	if [[ -n "$1" && "$1" =~ ^${IP_MATCH_REGEXP}$ ]]
	then	echo "Subject: abuse report about $1 - ${datefield}"
	else	echo "Subject: abuse report - ${datefield}"
	fi
	echo "Mime-Version: 1.0"
	echo -e "Content-Type: multipart/mixed; charset=utf8; boundary=\"${MIME_BOUNDARY}\"\n"
	echo -e "${mime_preamble}\n\n"
}

build_report_mail_first_mime_part_xarf_format() {
	# generate the first mime part of a x-arf mail
	# three args with ip address, counter, amount of reported ip addresses
	# returned value on stdout
	if [[ -n "$1" && "$1" =~ ^${IP_MATCH_REGEXP}$ ]]
	then	echo -e "--${MIME_BOUNDARY}\nMIME-Version: 1.0\nContent-Transfer-Encoding: 7bit"
		echo -e "Content-Type: text/plain; charset=utf8;\n\n"
		echo -en "Dear DFN-CERT,\n\nthis is an automated report "
		echo -n "for ip address $1"
		echo -en " in format \"X-ARF\" generated on "
		date +%F" "%H:%M:%S" "%z
		echo -e "\nip address ${ip} produced ${loglinecount[${ip}]} log lines, sample log lines attached."
		echo -e "\nRegards,\n${data_supplier_organisation}\n\n"
	else	echo -n "ERROR: you must specify an ip address, a counter, and the amount of reported"
		echo " ip addresses as argument for build_report_mail_first_mime_part_xarf_format."
		exit 1
	fi
}

build_report_ids() {
	# generate for all ip addresses the corresponding Report-ID
	# Report-ID should be "[UNIXTIMESTAMP|randomnumber(4)]@domain.tld"
	# using unixtimestamp of first occurence of incident, this is not specified accurately
	# no args, uses global vars
	# no return value
	for ip in ${!iphash[@]}
	do	first_datetime=$( echo -e "${iphash[$ip]}" | ${HEAD_BIN} -n 1 | ${CUT_BIN} -d" " -f1-3 )
		report_id=$( date +%s --date="${first_datetime}" )
		# get a random number of exactly 4 digits
		[[ "$RANDOM$RANDOM$RANDOM$RANDOM" =~ ([0-9]{4}) ]] && report_id+=${BASH_REMATCH[1]}
		ipreportid[${ip}]="${report_id}@${data_supplier_domain}"
	done
}

build_report_mail_second_mime_part_xarf_format() {
	# generate the second mime part of a x-arf mail
	# one arg with ip address
	# returned value on stdout
	if [[ -n "$1" && "$1" =~ ^${IP_MATCH_REGEXP}$ ]]
	then	ip="$1"
		echo -e "--${MIME_BOUNDARY}\nMIME-Version: 1.0\nContent-Transfer-Encoding: 7bit"
		echo -e "Content-Type: text/plain; charset=utf8; name=\"report.txt\";\n"
		echo "---"
		echo "Reported-From: ${reported_from_email_address}"
		echo "Category: abuse"
		echo "Report-Type: login-attack"
		echo "Service: ssh"
		echo "Port: 22"
		echo "User-Agent: ${SCRIPT_NAME} ${VERSION}"
		echo "Report-ID: ${ipreportid[${ip}]}"
		# first occurence of incident:
		firstseen=$(echo -e "${iphash[$ip]}" | ${HEAD_BIN} -n 1 | ${CUT_BIN} -d" " -f1-3)
		echo -n "Date: "
		# RFC 3339 or 2822???
		# date --rfc-3339=seconds --date="${firstseen}"
		date -R --date="${firstseen}"
		echo "Source: ${ip}"
		echo "Source-Type: ipv4"
		echo "Attachment: text/plain"
		echo "Schema-URL: http://www.x-arf.org/schema/abuse_login-attack_0.1.1.json"
		echo
	else	echo "ERROR: you must specify an ip address as argument for build_report_mail_second_mime_part_xarf_format."
		exit 1
	fi	
}

build_report_mail_third_mime_part_xarf_format() {
	# generate the third mime part of a x-arf mail
	# one arg with ip address
	# returned value on stdout
	if [[ -n "$1" && "$1" =~ ^${IP_MATCH_REGEXP}$ ]]
	then	ip="$1"
		echo -e "--${MIME_BOUNDARY}\nMIME-Version: 1.0\nContent-Transfer-Encoding: 7bit"
		echo -e "Content-Type: text/plain; charset=utf8; name=\"logfile.log\";\n"
		echo -e "${iphash[$ip]}\n"
	else	echo "ERROR: you must specify an ip address as argument for build_report_mail_third_mime_part_xarf_format."
		exit 1
	fi
}

send_reports_in_xarf_format() {
	# send abuse mails in x-arf format for every ip address
	# no args, uses global vars
	# no return value
	unique_ips=${#iphash[@]}
	ip_counter=1
	for ip in ${!iphash[@]}
	do	(	build_report_mail_header_xarf_format "${ip}"
			build_report_mail_first_mime_part_xarf_format "${ip}"
			build_report_mail_second_mime_part_xarf_format "${ip}"
			build_report_mail_third_mime_part_xarf_format "${ip}"
			# closing MIME parts
			echo "--${MIME_BOUNDARY}--"
		) | ${SENDMAIL_BIN} ${sendmail_args}
		ip_counter=$(($ip_counter + 1))
	done
}

send_status_mail() {
	# send out a mail with a short status report of sent abuse mails
	# one optional arg: status message
	# returned value on stdout
	if [ -n "${status_email_address}" ]
	then	(	echo "From: ${from_email_address}"
			echo "To: ${status_email_address}"
			echo "Reply-To: ${replyto_email_address}"
			echo "Organisation: ${data_supplier_organisation}"
			echo "X-System-Id: ${system_id}"
			echo "X-Script-Version: ${VERSION}"
			echo "X-Script-Name: ${SCRIPT_NAME}"
			echo "X-Mailer: ${SCRIPT_NAME} ${VERSION}"
			echo "Auto-Submitted: auto-generated"
			echo "Subject: ${SCRIPT_NAME}: auto-generated status report"
			echo -en "\nDear ${data_supplier_organisation},\n\nthis is an automated status report generated on "
			date +%F" "%H:%M:%S" "%z
			if [ -n "$1" ]
			then	echo -e "\n---- begin status message ----"
				echo -e "$1"
				echo -e "---- end status message ----"
			fi
			echo -e "\nRegards,\n${data_supplier_organisation}\n"
		) | ${SENDMAIL_BIN} ${sendmail_args}
	fi
}

main() {
	[[ "${remember_last_loglines}"x == "yes"x ]] && read_lastloglines
	read_and_parse_logfiles
	[[ "${remember_last_loglines}"x == "yes"x ]] && write_lastloglines
	build_iphash
	apply_threshold
	apply_cidr_blacklist
	# do we still have any ip addresses left to report?
	if [[ ${#iphash[@]} -gt 0 ]]
	then	sort_and_limit_reportable_loglines
		anonymize_log_content
		build_report_ids
		if [[ "${export_format}"x == "X-ARF"x ]]
		then	send_reports_in_xarf_format
		else	echo "ERROR: \"${export_format}\" is not a supported report format."
			exit 1
		fi
		status_text="reported ${#iphash[@]} unique ip addresses in ${export_format} format to ${report_email_address}"
		for ip in ${!iphash[@]}
		do	status_text+="\n${ip} with Report-ID ${ipreportid[${ip}]} produced ${loglinecount[${ip}]} loglines"
		done
		send_status_mail "${status_text}"
	elif	[[ "${send_always_status_mail}"x == "yes"x ]]
	then	send_status_mail "${status_message_if_nothing_was_reported}"
	fi
}
#####################
### end functions ###
#####################

main
exit 0

